/*-------------------------------------------------------------------*/
/* randompw.c                                                        */
/* ----------                                                        */
/*                                                                   */
/* Program file for "randompw".  "randompw" is a program that will   */
/* generate random passwords.  These password can be between four    */
/* and thirty two bytes (characters) in length.  Depending on the    */
/* options chosen, "randompw" will generate passwords using upper    */
/* case, lower case, numbers and symbols or any combination of the   */
/* latter.                                                           */
/*                                                                   */
/* "randompw" can be used within a "useradd" script to automatically */
/* generate passwords whenever a new user is defined.                */
/*                                                                   */
/* Copyright (c) 1987-2005                                           */
/* Marc Niegowski                                                    */
/* Connectivity, Inc.                                                */
/* All rights reserved.                                              */
/*                                                                   */
/* 23 W. Fourth Street                                               */
/* Media, PA 19063-2805                                              */
/* USA                                                               */
/*                                                                   */
/* Phone: (610) 566-0227                                             */
/* Fax:   (610) 566-0641                                             */
/*                                                                   */
/* Web:   http://www.marcsweb.com                                    */
/* FTP:   ftp://ftp.marcsweb.com                                     */
/*                                                                   */
/* Email: Marc@Tech-Center.com                                       */
/*-------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef _WIN32

#include <sys/timeb.h>

#else

#include <sys/time.h>
#include <unistd.h>

#endif

#include <randompw.h>

/*-------------------------------------------------------------------*/
/* main                                                              */
/* ----                                                              */
/*                                                                   */
/* "randompw" entry point.                                           */
/*                                                                   */
/* Call with: int    argc - Number of arguments.                     */
/*                                                                   */
/*            char **argv - A pointer to an array of character array */
/*                          arguments (parameters).                  */
/*                                                                   */
/*            char **envp - A pointer to an array of character array */
/*                          environment variables.                   */
/*                                                                   */
/* Returns:   int         - An integer return code.                  */
/*                          zero for success.  Non-zero indicates an */
/*                          error condition.                         */
/*-------------------------------------------------------------------*/

int
main(int argc, char **argv, char **envp)
{
    rpw_struct                         // Parameter structure for
    rpw_st;                            // internal randompw parameters.  

    rpw_getargs(&rpw_st, argc, argv);  // Process argument list.

    if (rpw_st.message)                // A returned message indicates
        return rpw_return(&rpw_st);    // an error or info request. 

    rpw_generate(&rpw_st);             // Generate a password.

    return RPW_NO_ERROR;               // Return success to caller.
}

/*-------------------------------------------------------------------*/
/* rpw_return                                                        */
/* ----------                                                        */
/*                                                                   */
/* Generic return routine to return a return code (zero for a suc-   */
/* cessful request, and non-zero for an unsuccessful request) and    */
/* produce a message (help text, version information or an error)    */
/* based upon the "message" pointer.                                 */
/*                                                                   */
/* Call with: rpw_struct  - A pointer to an rpw_struct structure     */
/*            *rpw_st       that contains a pointer to a character   */
/*                          array message and an integer return      */
/*                          return code.                             */
/*                                                                   */
/* Returns:   int         - The integer return code that was passed  */
/*                          in the rpw_struct structure.             */
/*-------------------------------------------------------------------*/

int
rpw_return(rpw_struct *rpw_st)
{
    printf(rpw_header_str);            // Display a common header.
    printf(rpw_st->message);           // Display the message.
    return rpw_st->retcode;            // Return the return code.
}

/*-------------------------------------------------------------------*/
/* rpw_getargs                                                       */
/* -----------                                                       */
/*                                                                   */
/* Process the arguments (parameters) that were passed to randompw   */
/* when it was invoked.  Populates the rpw_struct structure depend-  */
/* ing upon the callers parameter options.                           */
/*                                                                   */
/* Call with: rpw_struct  - A pointer to an empty rpw_struct struct- */
/*            *rpw_st       ure.                                     */
/*                                                                   */
/*            int    argc - An integer containing the number of arg- */
/*                          uments passed by the caller.             */
/*                                                                   */
/*            char **argv - A pointer to an array of character arg-  */
/*                          uments passed by the caller.             */
/*                                                                   */
/* Returns:   void        - Nothing.                                 */
/*-------------------------------------------------------------------*/

void
rpw_getargs(rpw_struct *rpw_st, int argc, char **argv)
{
    int   ix;

    rpw_st->pwlen    = RPW_DEFAULT_LENGTH;
    rpw_st->pwcount  = RPW_DEFAULT_COUNT;
    rpw_st->flags    = RPW_CLEAR_FLAG;
    rpw_st->message  = 0;
    rpw_st->retcode  = RPW_NO_ERROR;

    if (argc == 1)
        return;

    for (ix = 1; ix < argc; ix++)
    {
        if (strlen(*(argv + ix)) < 2)
        {
            rpw_parm_error(rpw_st);
            return;
        }

        if (**(argv + ix) == '-')
        {
            if (*(*(argv + ix) + 1) == '-')
            {
                if (rpw_long_parm(rpw_st, *(argv + ix) + 2) ==
                           RPW_PARM_ERROR)
                    return;
                else
                    continue;
            }
            else
                if (rpw_short_parm(rpw_st, *(argv + ix) + 1) ==
                           RPW_PARM_ERROR)
                    return;
        }
        else
        {
            rpw_parm_error(rpw_st);
            return;
        }
    }

    return;
}

/*-------------------------------------------------------------------*/
/* rpw_short_parm                                                    */
/* --------------                                                    */
/*                                                                   */
/* Process the short parameters (i.e., a hyphen '-'  followed by one */
/* character).   Populates the rpw_struct structure depending on the */
/* callers parameter options.   Multiple short parameters may follow */
/* the hyphen.                                                       */
/*                                                                   */
/* Call with: rpw_struct  - A pointer to an empty rpw_struct struct- */
/*            *rpw_st       ure.                                     */
/*                                                                   */
/*            char    *pc - A pointer to a string of one or more     */
/*                          short parameters.                        */
/*                                                                   */
/* Returns:   int         - An integer return code, RPW_NO_ERROR (0) */
/*                          or RPW_PARM_ERROR (-1).                  */
/*-------------------------------------------------------------------*/

int
rpw_short_parm(rpw_struct *rpw_st, char *pc)
{
    while (*pc)
    {
        switch (*pc++)
        {
            case 'h':
                rpw_st->message = (char *) rpw_help_str;
                break;
            case 'v':
                if (!(rpw_st->message))
                    rpw_st->message = (char *) rpw_version_str;
                break;
            case 's':
                if (!(rpw_st->pwlen =
                      rpw_check_limit(pc, RPW_MIN_LENGTH, RPW_MAX_LENGTH)))
                    return rpw_parm_error(rpw_st);

                pc += strspn(pc, "0123456789");

                break;
            case 'n':
                if (!(rpw_st->pwcount =
                      rpw_check_limit(pc, RPW_MIN_COUNT, RPW_MAX_COUNT)))
                    return rpw_parm_error(rpw_st);

                pc += strspn(pc, "0123456789");

                break;
            case 'u':
                if (rpw_st->flags & RPW_LOWER_FLAG)
                {
                    rpw_st->message = (char *) rpw_logic_str;
                    rpw_st->retcode = RPW_PARM_ERROR;
                    return RPW_PARM_ERROR;
                }

                rpw_st->flags |= RPW_UPPER_FLAG;
                break;
            case 'l':
                if (rpw_st->flags & RPW_UPPER_FLAG)
                {
                    rpw_st->message = (char *) rpw_logic_str;
                    rpw_st->retcode = RPW_PARM_ERROR;
                    return RPW_PARM_ERROR;
                }

                rpw_st->flags |= RPW_LOWER_FLAG;
                break;
            case 'a':
                if (rpw_st->flags & RPW_NUMBER_FLAG)
                {
                    rpw_st->message = (char *) rpw_rules_str;
                    rpw_st->retcode = RPW_PARM_ERROR;
                    return RPW_PARM_ERROR;
                }

                rpw_st->flags |= RPW_ALPHA_FLAG;
                break;
            case 'r':
                if (rpw_st->flags & RPW_ALPHA_FLAG)
                {
                    rpw_st->message = (char *) rpw_rules_str;
                    rpw_st->retcode = RPW_PARM_ERROR;
                    return RPW_PARM_ERROR;
                }

                rpw_st->flags |= RPW_NUMBER_FLAG;
                break;
            case 'y':
                rpw_st->flags |= RPW_SYMBOL_FLAG;
                break;
            case 'q':
                rpw_st->flags |= RPW_SEQUENCE_FLAG;
                break;
            default:
                return rpw_parm_error(rpw_st);
        }
    }

    return RPW_NO_ERROR;
}

/*-------------------------------------------------------------------*/
/* rpw_check_limit                                                   */
/* ---------------                                                   */
/*                                                                   */
/* This function is called to test whether a parameter with a num-   */
/* eric argument falls within a specific range.                      */
/*                                                                   */
/* Call with: char    *pc - A pointer to a string of numbers that    */
/*                          may be followed by other characters.     */
/*                                                                   */
/*            int    imin - An integer minimum number.               */
/*                                                                   */
/*            int    imax - An integer maximum number.               */
/*                                                                   */
/* Returns:   int         - FALSE (0) if the string was not numeric  */
/*                          or the number was outside the allowable  */
/*                          range, or the integer value of the num-  */
/*                          eric string.                             */
/*-------------------------------------------------------------------*/

int
rpw_check_limit(char *pc, int imin, int imax)
{
    int    ix;

    if (!(*pc))
        return FALSE;

    ix = atoi(pc);

    if ((ix < imin) ||
        (ix > imax))
       return FALSE;

    return ix;
}

/*-------------------------------------------------------------------*/
/* rpw_parm_error                                                    */
/* --------------                                                    */
/*                                                                   */
/* Called by rpw_getargs, rpw_short_parm and rpw_long parm to popul- */
/* ate the rpw_struct structure with an error message and an error   */
/* code.                                                             */
/*                                                                   */
/* Call with: rpw_struct  - A pointer to an empty rpw_struct struct- */
/*            *rpw_st       ure.                                     */
/*                                                                   */
/* Returns:   int         - An integer RPW_PARM_ERROR (-1 always)    */
/*                          since this function is only called upon  */
/*                          detection of a parameter error.          */
/*-------------------------------------------------------------------*/

int
rpw_parm_error(rpw_struct *rpw_st)
{
    rpw_st->message = (char *) rpw_error_str;
    rpw_st->retcode = RPW_PARM_ERROR;
    return RPW_PARM_ERROR;
}

/*-------------------------------------------------------------------*/
/* rpw_long_parm                                                     */
/* -------------                                                     */
/*                                                                   */
/* Process the long parameters (i.e., a double hyphen '--' followed  */
/* by a parameter word).  Populates the rpw_struct structure depend- */
/* ing on the callers parameter options.  A double hyphen '--' must  */
/* only be followed by a single parameter word.                      */
/*                                                                   */
/* Call with: rpw_struct  - A pointer to an empty rpw_struct struct- */
/*            *rpw_st       ure.                                     */
/*                                                                   */
/*            char    *pc - A pointer to a string of one long param- */
/*                          eter.                                    */
/*                                                                   */
/* Returns:   int         - An integer return code, RPW_NO_ERROR (0) */
/*                          or RPW_PARM_ERROR (-1).                  */
/*-------------------------------------------------------------------*/

int
rpw_long_parm(rpw_struct *rpw_st, char *pc)
{
    int   ix, il;

    char  ps[4];
    char  pn[3];

    *ps = 0;
    *pn = 0;

    if ((il = strcspn(pc, "123456789")))
        if (atoi(pc + il))
        {
            if (strlen(pc + il) > (sizeof(pn) - 1))
                return rpw_parm_error(rpw_st);
            else
                strcpy(pn, pc + il);
        }

    il = strcspn(pc, ":=");

    for (ix = 0; ix < (sizeof(rpw_parm) / sizeof(*rpw_parm)); ix++)
         if (!(strncmp(pc, *(*(rpw_parm + ix)), il)))
        {
             strcpy(ps, *(*(rpw_parm + ix) + 1));
             strcat(ps, pn);
             return rpw_short_parm(rpw_st, ps);
        }
    return rpw_parm_error(rpw_st);
}

/*-------------------------------------------------------------------*/
/* rpw_generate                                                      */
/* ------------                                                      */
/*                                                                   */
/* Generate a random password using the random number generator and  */
/* rules based upon the option flags found in the rpw_struct struct- */
/* ure.                                                              */
/*                                                                   */
/* Call with: rpw_struct  - A pointer to an rpw_struct structure     */
/*            *rpw_st       populated with option flags.             */
/*                                                                   */
/* Returns:   int         - An integer return code, RPW_NO_ERROR (0) */
/*                          or RPW_TIME_ERROR (-2).                  */
/*-------------------------------------------------------------------*/

int
rpw_generate(rpw_struct *rpw_st)
{
    rpw_tv tv;

    int    ix;

#ifdef _WIN32
    unsigned
    short  sd;
#else
    long
    int    sd;
#endif

    char   ptab[(RPW_ALPHA_SIZE  * 2) +
                 RPW_NUMBER_SIZE      +
                 RPW_SYMBOL_SIZE + 1],
           pstr[RPW_MAX_LENGTH   + 1],
           pchr[2];

    char  *px;

    float  f;

    f = 0.0 + RPW_ALPHA_SIZE;

    if (rpw_st->flags & RPW_UPPER_FLAG)
        strcpy(ptab, rpw_pwupper);
    else
    {
        if (rpw_st->flags & RPW_LOWER_FLAG)
            strcpy(ptab, rpw_pwlower);
        else
        {
            strcpy(ptab, rpw_pwupper);
            strcat(ptab, rpw_pwlower);
            f *= 2.0;
        }
    }

    if (!(rpw_st->flags & RPW_ALPHA_FLAG))
    {
        strcat(ptab, rpw_pwnumbers);
        f += RPW_NUMBER_SIZE;
    }

    if (rpw_st->flags & RPW_SYMBOL_FLAG)
    {
        strcat(ptab, rpw_pwsymbols);
        f += RPW_SYMBOL_SIZE;
    }

#ifdef _WIN32
    _ftime(&tv);

    sd = tv.millitm;
#else
    if (gettimeofday(&tv, 0) == -1)
        return RPW_TIME_ERROR;

    sd = tv.tv_usec;
#endif

    srand((time(NULL) * 1000) + sd);

    while (rpw_st->pwcount--)
    {
        *pstr       = 0;
        *(pchr + 1) = 0;

        for (ix = 0; ix < rpw_st->pwlen; ix++)
        {
            while (TRUE)
            {
                *pchr = ptab[(int) (f * rand() / (RAND_MAX + 1.0))];

                if ((ix) && (rpw_st->flags & RPW_SEQUENCE_FLAG))
                {
                    px = pstr + ix - 1;

                    if ((*px == *pchr)     ||
                        (*px == *pchr - 1) ||
                        (*px == *pchr + 1))
                        continue;
                }

                if (*pchr >= '0' && *pchr <= '9')
                    rpw_st->flags |= RPW_HASNUM_FLAG;

                if ((*pchr >= 'a' && *pchr <= 'z') ||
                    (*pchr >= 'A' && *pchr <= 'Z'))
                    rpw_st->flags |= RPW_HASALPHA_FLAG;
  
                strcat(pstr, pchr);
                break;
            }
        }

        if (rpw_st->flags & RPW_NUMBER_FLAG)
            if (!(rpw_st->flags & RPW_HASNUM_FLAG))
                rpw_inject_char(pstr, (char *) rpw_pwnumbers);

        // This is highly unlikely, but it would be a defect if I didn't
        // handle a situation where no alpha characters are generated in
        // a password (the shorter the password, the more likely - but 
        // during testing, I couldn;t even get 32,767 four character
        // passwords to generate without an alpha character.

        if (!(rpw_st->flags & RPW_HASALPHA_FLAG))
        {
            if (!(rpw_st->flags & RPW_LOWER_FLAG))
                rpw_inject_char(pstr, (char *) rpw_pwupper);
            else
                rpw_inject_char(pstr, (char *) rpw_pwlower);
        }

        printf("%s\n", pstr);

        rpw_st->flags &= ~(RPW_HASNUM_FLAG | RPW_HASALPHA_FLAG);
    }

    return RPW_NO_ERROR;
}

/*-------------------------------------------------------------------*/
/* rpw_inject_char                                                   */
/* ---------------                                                   */
/*                                                                   */
/* Called by rpw_generate to 'inject' at least one character to en-  */
/* force the 'at least one number' or the 'at least one alpha' rule. */
/* This function will force at least one character from the passed   */
/* string pointer into the generated password.                       */
/*                                                                   */
/* Call with: char *pstr  - A pointer to the password string.        */
/*                                                                   */
/*            char *ptab  - A pointer to a string containing the     */
/*                          list of characters from which to pick.   */
/*                                                                   */
/* Returns:   void        - Nothing.                                 */
/*-------------------------------------------------------------------*/

void
rpw_inject_char(char *pstr, char *ptab)
{
    char   c;
    int    i;
    float  f;

    f = (float) strlen(ptab);        
    c = ptab[(int) (f * rand() / (RAND_MAX + 1.0))];
    f = (float) strlen(pstr);
    i = (int) (f * rand() / (RAND_MAX + 1.0));

    pstr[i] = c;
}
